/*
 * Ping.java 
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.veraxsystems.icmp.ping;

import com.veraxsystems.icmp.ping.handlers.Sender;
import com.veraxsystems.icmp.ping.handlers.Receiver;
import com.veraxsystems.icmp.ping.handlers.PingRequest;
import com.veraxsystems.icmp.ping.exception.PingErrorException;
import com.veraxsystems.icmp.jni.IcmpJniBridge;
import com.veraxsystems.icmp.jni.exception.IcmpSystemException;
import com.veraxsystems.icmp.ping.handlers.BlockingPing;
import java.net.InetAddress;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * Ping is the class implementing the ICMP Echo request known as network "ping" functionality.
 * Class is defined as static and requires initialization before it can be used {@link #init}.
 * The ICMP communication is performed via {@link IcmpJniBridge} object that is created during 
 * initialization (only one such object is used for all echo requests). 
 * Class uses two internal threads for its activity. One reads ICMP application issued requests and 
 * sends  ICMP echo messages while the second one reads responses and handles requests timeouts.
 * At the end of using the {@link #release} method must be called to release resources especially
 * ICMP socket.
 * ICMP echo requests (pings) are sent asynchronously. Requests are issued by {@link #echo echo()}
 * method that has provided result handler executed either on request response or processing error.
 * Obtaining error as result of operation means that either fatal error occurred or Ping class has been
 * released. In such case further processing is not possible and Ping class must be re-initialized.
 * All methods are synchronized to avoid sending echo during class initialization/releasing.
 */
public class Ping 
{
   private static IcmpJniBridge icmp = null;
   private static Sender sender = null;
   private static Receiver receiver = null;
   private static Thread senderThread = null;
   private static Thread receiverThread = null;
   
   
   /**
    * Initialize Ping class. Three is created ICMP socket and started two processing threads - one for
    * sending ICMP requests and second for receiving responses and timeouts handling.
    * To release resources the {@link #release} method must be used. Without it allocated
    * resources are not released.
    * @throws PingErrorException issued when either initialization already has been done or 
    * other fatal error occurs (with socket creation or thread starting)
    */
   static public synchronized void init() throws PingErrorException
   {
      if (icmp == null)
      {
         try
         {
            icmp = new IcmpJniBridge();
            icmp.openSocket();
            try
            {
               receiver = new Receiver(icmp);
               sender = new Sender(receiver, icmp);
               receiverThread = new Thread(receiver);
               senderThread = new Thread(sender);
               receiverThread.start();
               senderThread.start();
            }
            catch(Exception ex)
            {
               release();
               throw new PingErrorException("Cannot initialize PingManager", ex);
            }
         }
         catch(IcmpSystemException ex)
         {
            icmp = null;
            throw new PingErrorException("Cannot open ICMP socket", ex);
         }
      }
      else throw new PingErrorException("Already initialized", null);
   }


   /**
    * Release resources allocated in {@link #init} method.
    * Method always succeeds. Eventual internal exceptions are either properly handled
    * or logged using {@link Logger}.
    */
   public static synchronized void release() 
   {
      try 
      {
         if (icmp != null) icmp.closeSocket();
      }
      catch (IcmpSystemException ex) 
      {
         Logger.getLogger(Ping.class.getName()).log(Level.SEVERE, "Socket close failed", ex);
      };
      try 
      {
         if (senderThread != null)
         {
            senderThread.interrupt();
            senderThread.join();
         }
         if (receiverThread != null)
         {
            receiverThread.interrupt();
            receiverThread.join();
         }
      } 
      catch (InterruptedException ex1) 
      {
         Logger.getLogger(Ping.class.getName()).log(Level.SEVERE, "Could not stop threads", ex1);
      }
      sender = null;
      receiver = null;
      senderThread = null;
      receiverThread = null;
      icmp = null; 
      return;
   }

   
   /**
    * Send ICMP echo request. Operation is executed asynchronously and result is propagated using handler object. 
    * Because handled is invoked by separate thread (receiving thread) a handler must be properly synchronized.
    * Ping class guarantees that for each request sent the handler will be called either on 
    * received response, on timeout or in case of error.
    * The only error (exception) reported by method is issued when Ping class is not initialized.
    * @param address IP address of host to send ICMP echo request
    * @param dataToSend data to be send as part of ICMP message
    * @param timeoutVal expressed in milliseconds time to wait for echo response
    * @param handler asynchronous result handler
    * @return identifier of echo message to be sent (the unique value on computer)
    * @throws PingErrorException issued if Ping class is not initialized 
    */
   public static synchronized int echo(InetAddress address, byte[] dataToSend, int timeoutVal, PingResponseHandler handler) throws PingErrorException
   {
      int messageId = 0;
      if (icmp != null)
      {
         messageId = icmp.getId();
         sender.registerRequest(new PingRequest(address, dataToSend, dataToSend.length, timeoutVal, messageId, handler));
      }
      else throw new PingErrorException("PingEngine not initialized", null);
      return messageId;
   }


   /**
    * Send default ICMP echo request. Operation is executed asynchronously and result is propagated using handler object. 
    * Because handled is invoked by separate thread (receiving thread) a handler must be properly synchronized.
    * Ping class guarantees that for each request sent the handler will be called either on 
    * received response, on timeout or in case of error.
    * The only error (exception) reported by method is issued when Ping class is not initialized.
    * Default data construction is based on Windows ping command behavior and contains string: "abc...w".
    * @param address IP address of host to send ICMP echo request
    * @param dataLength the length of default data to be sent
    * @param timeoutVal expressed in milliseconds time to wait for echo response
    * @param handler asynchronous result handler
    * @return identifier of echo message to be sent (the unique value on computer)
    * @throws PingErrorException issued if Ping class is not initialized 
    */
   public static synchronized int echo(InetAddress address, int dataLength, int timeoutVal, PingResponseHandler handler) throws PingErrorException
   {
      int messageId = icmp.getId();
      if (sender != null)
         sender.registerRequest(new PingRequest(address, null, dataLength, timeoutVal, messageId, handler));
      else
         throw new PingErrorException("PingEngine not initialized", null);
      return messageId;
   }


   /**
    * Send the set of ICMP echo requests. Method is executed synchronously and waits until all echo requests
    * are finished. Internally, echo requests are send asynchronously - it means all requests are send immediately
    * without waiting for responses (those, are handled in separate thread). Next, method waits until either all echo 
    * responses are received or timeouts occurs.
    * <p>
    * Each echo result is placed in array of results corresponding to <code>addresses</code> parameter (it means for each input host 
    * the result is placed in corresponding element of result array). 
    * <p>
    * Each request timeout is calculated separately - it means that its calculation starts in the moment
    * when request is sent (not when <code>echo</code> method is called). Such behavior leads to 
    * situation that <code>echo</code> execution lasts slightly longer than timeout specified.
    * <p>
    * Method reports exception when either {@link Ping} is not initialized or current thread is interrupted.
    * @param addresses array of host addresses send ICMP echo requests
    * @param dataToSend data to be send as part of ICMP message
    * @param timeout expressed in milliseconds time to wait for echo response
    * @return the array of results
    * @throws PingErrorException issued if Ping class is not initialized or thread is interrupted  
    */
   public static PingResult[] echo(InetAddress[] addresses, byte[] dataToSend, int timeout) throws PingErrorException 
   {
      BlockingPing ping = new BlockingPing();
      return ping.echo(addresses, dataToSend, dataToSend.length, timeout);
   }

   /**
    * Send the set of default ICMP echo requests. Method is executed synchronously and waits until all echo requests
    * are finished. Internally, echo requests are send asynchronously - it means all requests are send immediately
    * without waiting for responses (those, are handled in separate thread). Next, method waits until either all echo 
    * responses are received or timeouts occurs.
    * <p>
    * Each echo result is placed in array of results corresponding to <code>addresses</code> parameter (it means for each input host 
    * the result is placed in corresponding element of result array). 
    * <p>
    * Each request timeout is calculated separately - it means that its calculation starts in the moment
    * when request is sent (not when <code>echo</code> method is called). Such behavior leads to 
    * situation that <code>echo</code> execution lasts slightly longer than timeout specified.
    * <p>
    * Method reports exception when either {@link Ping} is not initialized or current thread is interrupted.
    * Default data construction is based on Windows ping command behavior and contains string: "abc...w".
    * @param addresses array of host addresses send ICMP echo requests
    * @param dataLength the length of default data to be sent
    * @param timeout expressed in milliseconds time to wait for echo response
    * @return the array of results
    * @throws PingErrorException issued if Ping class is not initialized or thread is interrupted  
    */
   public static PingResult[] echo(InetAddress[] addresses, int dataLength, int timeout) throws PingErrorException 
   {
      BlockingPing ping = new BlockingPing();
      return ping.echo(addresses, null, dataLength, timeout);
   }

}
