/*
 * Receiver.java 
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.veraxsystems.icmp.ping.handlers;

import com.veraxsystems.icmp.IcmpStatus;
import com.veraxsystems.icmp.jni.IcmpEchoResponse;
import com.veraxsystems.icmp.jni.IcmpJniBridge;
import com.veraxsystems.icmp.jni.exception.IcmpException;
import com.veraxsystems.icmp.jni.exception.IcmpTimeoutException;
import com.veraxsystems.icmp.ping.Ping;
import com.veraxsystems.icmp.ping.exception.PingErrorException;
import com.veraxsystems.icmp.util.TimeUtil;

import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;


/**
 * An ICMP receiving thread implementation. 
 * The class is a {link IcmpJniBridge} receiver used in {@link Ping} class.
 * It handles list of ICMP requests that are registered by {@link Receiver#registerRequest registerRequest} method.
 * It receives ICMP responses and for those ones that matches registered requests, executes request result handler.
 * The requests timeout condition is handled as well, so timeout handler is executed for requests that passes their end time.
 */
public class Receiver implements Runnable 
{
   private IcmpJniBridge icmp = null;
   private Queue<PingRequest> requestsQueue = null;
   private Lock mutex = null;
   private Condition hasData = null;

   
   /**
    * Constructor accepts <code>IcmpJniBridge</code> object that is used to receive ICMP messages.
    * @param icmpVal ICMP socket
    */
   public Receiver(IcmpJniBridge icmpVal) 
   {
      icmp = icmpVal;
      requestsQueue = new PriorityQueue<PingRequest>();
      mutex = new ReentrantLock();
      hasData = mutex.newCondition();
   }

   /*
    * @brief Validate (filter) ICMP response.
    * Method checks if ICMP response matches ICMP Echo Request parameters. Such filtering is
    * needed because ICMP socket receives all ICMP messages accepted by network interfaces.
    */
   private boolean validResponse(PingRequest request, IcmpEchoResponse responseMsg) 
   {
      boolean result;
      
      result = responseMsg.getStatus() != IcmpStatus.UnexpectedResponse;
      if (result) result = responseMsg.getMessageId() == request.getMessageId();
      if (result) result = responseMsg.getDestinationAddress().equals(request.getAddress());
      return result;
   }


   /**
    * Register ICMP EchoRequest and put it into requests queue. Because the queue is sorted by 
    * <i>end time</i> so the first request is the one with the soonest timeout condition.
    * This feature is used to handle request timeouts in proper order.
    * @param request the ICMP descriptor
    * @throws PingErrorException exception is thrown if <code>Receiver</code> object is disposed
    */
   public void registerRequest(PingRequest request) throws PingErrorException 
   {
      mutex.lock();
      try 
      {
         if (requestsQueue != null) 
         {
            PingRequest requestCopy = new PingRequest(request);
            requestsQueue.add(requestCopy);
            hasData.signal();
         }
         else throw new PingErrorException("Not initialized", null);
      }
      finally 
      {
         mutex.unlock();
      };
      return;
   }

   
   /*
    * Main activity of <code>Receiver</code> thread.
    */
   public void run() 
   {
      IcmpEchoResponse responseMsg = null;
      int timeoutTime;
      long endTime;
      
      try
      {
         for (;;) 
         {
            try 
            {
               // Take smallest end time value to handle timeout 
               endTime = getMinimumEndTime();
               timeoutTime = (int) (endTime - TimeUtil.currentTimeMillis());
               
               // Wait for message or handle timeout (via exception)
               icmp.selectSocket(timeoutTime);
               
               // Receive ICMP message
               responseMsg = icmp.receive();
               
               // Handle received response
               handleResponse(responseMsg);
            }
            catch (IcmpTimeoutException ex) 
            {
               handleTimeout();
            }
         }
      } 
      catch (IcmpException ex) 
      {
         cleanupRequestsQueue(ex);
      } 
      catch (InterruptedException ex) 
      {
         cleanupRequestsQueue(ex);
      }
      finally 
      {
         dispose();
      }
   }

   /*
    * Handle timeout condition. Timeout occurs always for first request in queue so
    * forst request is taken and its timeout handler is executed
    */
   private void handleTimeout() 
   {
      PingRequest request;
      
      mutex.lock();
      try 
      {
         request = requestsQueue.poll();
         request.getHandler().handleTimeout(request);
      }
      finally 
      {
         mutex.unlock();
      };
      return;
   }

   /*
    * Handle received ISMP response. The responses order is unknown, so
    * find appropriate request and execute its response handler. The handled
    * request is removed from queue.
    */
   private void handleResponse(IcmpEchoResponse responseMsg) 
   {
      PingRequest request;
      Iterator<PingRequest> it;
      
      mutex.lock();
      try 
      {
         it = requestsQueue.iterator();
         while (it.hasNext()) 
         {
            request = it.next();
            if (validResponse(request, responseMsg)) 
            {
               request.getHandler().handleResponse(request, responseMsg);
               it.remove();
            }
         }
      } 
      finally 
      {
         mutex.unlock();
      }
      return;
   }

   
   /*
    * In case of socket error, whole processing is terminated and request queue is emptied.
    * All no handler requests have executed exception handler.
    */
   private void cleanupRequestsQueue(Exception ex) 
   {
      PingRequest request;
      
      mutex.lock();
      try
      {
         request = requestsQueue.poll();
         while (request != null) 
         {
            request.getHandler().handleException(request, new PingErrorException("ICMP socket fatal error", ex));
            request = requestsQueue.poll();
         };
         requestsQueue = null;
      } 
      finally 
      {
         mutex.unlock();
      }
      return;
   }

   /*
    * Get end time of request that will be timeouted first (it is first one in sorted queue). 
    * If there is no reuest in queue, operation waits until queue is not empty.
    */
   private long getMinimumEndTime() throws InterruptedException 
   {
      long result;
      
      mutex.lock();
      try
      {
         while (requestsQueue.size() == 0) 
         {
            hasData.await();
         }
         result = requestsQueue.peek().getEndTime();
      } 
      finally 
      {
         mutex.unlock();
      };
      return result;
   }

   
   /*
    * Simple resources releasing
    */
   private void dispose() 
   {
      icmp = null;
      requestsQueue = null;
      hasData = null;
      mutex = null;
      return;
   }
   
}
