/*
 * IcmpJniBridgeTest.java 
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.veraxsystems.icmp.jni;

import com.veraxsystems.icmp.jni.IcmpJniBridge;
import com.veraxsystems.icmp.jni.exception.IcmpFatalErrorException;
import com.veraxsystems.icmp.jni.exception.IcmpSystemException;
import com.veraxsystems.icmp.jni.exception.IcmpTimeoutException;
import com.veraxsystems.icmp.jni.handlers.Counter;
import com.veraxsystems.icmp.jni.handlers.PingLoopTest;
import com.veraxsystems.icmp.jni.handlers.ReceiveHandler;
import com.veraxsystems.icmp.jni.handlers.SendHandler;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Unit test for ICMP JNI interface.
 */
public class IcmpJniBridgeTest extends TestCase
{
    
   /**
    * Loads native library.
    */
   static {
       com.veraxsystems.icmp.jni.IcmpJniBridge.loadNativeLibrary();
   }
   
   /**
    * Create the test case
    *
    * @param testName name of the test case
    */
   public IcmpJniBridgeTest( String testName )
    {
       super(testName );
    }
   
   /**
    * @return the suite of tests being tested
    */
   public static Test suite()
   {
      return new TestSuite(IcmpJniBridgeTest.class);
   }
   
   
   /**
    * Simple test for ICMP JNI library loading. 
    */
   public void testLoadApi()
   {
      System.out.println("testLoadApi");
      IcmpJniBridge icmp = new IcmpJniBridge();
      icmp = null;
      assertTrue( true );
    }
   
   
 
   /**
    * Test basing ICMP JNI send/receive. The IcmpJniBridge object is
    * created, ICMP echo message is send success is achieved when
    * either response is received or timeout occurs (no API errors are allowed)
    */
   public void testBasicSendReceive()
   {
      System.out.println("testBasicSendReceive");
      try {
         IcmpJniBridge icmp = new IcmpJniBridge();
         Inet4Address address = (Inet4Address)InetAddress.getByName("8.8.8.8");
         icmp.openSocket();
         icmp.send(address, 1, "testBasicSendReceive".getBytes(), "testBasicSendReceive".getBytes().length);
         icmp.selectSocket(10000);
         icmp.receive();
         icmp.closeSocket();
         icmp = null;
         assertTrue( true );
      }
      catch (IcmpFatalErrorException ex) 
      {
         assertTrue( false );
      } 
      catch (IcmpTimeoutException ex) 
      {
         assertTrue( true );
      } 
      catch (UnknownHostException ex) 
      {
         assertTrue( false );
      } 
      catch (IcmpSystemException ex) 
      {
         ex.printStackTrace();
         assertTrue( false );
      }
   }
   
   
   /**
    * Test using many ICMP socket in parallel threads. Each thread performs socket operations
    * in polling mode - it means single thread is used to send data and receive responses.
    * No API errors are allowed here.
    */
   public void testThreads()
   {
      System.out.println("testThreads - rum polling loops in many threads");
      ExecutorService pool;
      Counter successCounter = new Counter();
      Counter failureCounter = new Counter();
      
      pool = Executors.newFixedThreadPool(200);
      for(int i = 0; i < 1; i++)
      {
         pool.execute(new PingLoopTest(40, successCounter, failureCounter));
      };
      pool.shutdown();
      try 
      {
         while(! pool.isTerminated())
         {
            Thread.sleep(1000);
         }
         pool.awaitTermination(240, TimeUnit.SECONDS);
         System.out.println("Success count: "+ successCounter.getCount());
         System.out.println("Failure count: "+ failureCounter.getCount());
         assertTrue(failureCounter.getCount() == 0);
      } catch (InterruptedException ex) 
      {
         assertTrue(false);
      }
   }
   
   /**
    * Test ICMP JNI interface by sending and receiving data on single IcmpJniBridge
    * object using two threads (sending and receiving threads). Again, no API
    * errors are allowed.
    */
   public void testSyncThreads()
   {
      System.out.println("testSyncThreads");
      final int echoNumber = 4000;
      Counter sentCounter;
      Counter handledCounter;
      IcmpJniBridge icmp = null;
      SendHandler sendHandler;
      ReceiveHandler receiveHandler;
      Thread sendThread = null;
      Thread receiveThread = null;
      
      try {
         sentCounter = new Counter();
         handledCounter = new Counter();
         icmp = new IcmpJniBridge();
         
         icmp.openSocket();
         sendHandler = new SendHandler(icmp, echoNumber, sentCounter);
         receiveHandler = new ReceiveHandler(icmp, echoNumber, handledCounter);
         
         sendThread = new Thread(sendHandler);
         receiveThread = new Thread(receiveHandler);
         
         receiveThread.start();
         sendThread.start();
         
         sendThread.join();
         sendThread = null;

         receiveThread.join();
         receiveThread = null;
         
         System.out.println("send count: "+ sentCounter.getCount());
         System.out.println("handled count: "+ handledCounter.getCount());
         assertTrue(true);
      }
      catch (InterruptedException ex) 
      {
         ex.printStackTrace();
         assertTrue(false);
      } 
      catch (IcmpSystemException ex) 
      {
         ex.printStackTrace();
         assertTrue(false);
      }
      finally
      {
         try 
         {
            if (icmp != null) icmp.closeSocket(); // This immediately stops threads so they may be joined 
         }
         catch (IcmpSystemException ex) 
         {
            // Ignore it - anyway processing is finished
         };
         try 
         {
            if (sendThread != null) sendThread.join();
            if (receiveThread != null) receiveThread.join();
         }
         catch (InterruptedException ex) 
         {
            ex.printStackTrace();
         }
      }
   }
   

}
