/*
 * PingTest.java 
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.veraxsystems.icmp.ping;

import com.veraxsystems.icmp.jni.IcmpEchoResponse;
import com.veraxsystems.icmp.jni.IcmpJniBridge;
import com.veraxsystems.icmp.ping.Ping;
import com.veraxsystems.icmp.ping.PingResponseHandler;
import com.veraxsystems.icmp.ping.PingResult;
import com.veraxsystems.icmp.ping.exception.PingErrorException;
import com.veraxsystems.icmp.ping.handlers.PingRequest;
import com.veraxsystems.icmp.util.TimeUtil;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Unit test Ping class
 */
public class PingTest extends TestCase 
{
    
    /**
     * Loads native library.
     */
    static {
        com.veraxsystems.icmp.jni.IcmpJniBridge.loadNativeLibrary();
    }    
   
   private final String testData = "PingTest: DATA";
   private final int testDefaultDataLen = 128;
   private final String testHosts[] = { 
           "127.0.0.1",
           "8.8.8.8",
           "1.2.3.4",
           "192.168.9.1",
           "news.yahoo.com",
           "www.google.com",
           "8.8.8.8",
           "www.vertel.com",
           "www.cnn.com",
           "www.foxnews.com",
           "www.accuweather.com",
           "www.huffingtonpost.com",
           "en.wikipedia.org",
           "www.hostitwise.com",
           "www.advancedgaragedoorinc.com",
           "www.bidobee.com",
           "www.imedicileather.com",
           "www.rushrash.com",
           "news.bbc.co.uk",
           "www.usatoday.com",
           "www.msnbc.com",
           "www.ikea.com",
           "www.veraxsystems.com",
           "10.0.0.0"
   };
   private InetAddress[] destinationAddresses;
   
   
   class ResponseHandler implements PingResponseHandler
   {
      CountDownLatch counter = null; 
      int responsesNumber;

      public ResponseHandler(CountDownLatch counterVal) 
      {
         counter = counterVal;
         responsesNumber = 0;
         return;
      }
      
      public void handleResponse(PingRequest request, IcmpEchoResponse response) 
      {
         System.out.println("Got response for address " + request.getAddress().getHostAddress() + " from " + response.getSourceAddress().getHostAddress() + " Status: " + response.getStatus() + " id: " + request.getMessageId() + " time:" + (TimeUtil.currentTimeMillis() - request.getSendTime()));
         counter.countDown();
         responsesNumber += 1;
      }
      
      public void handleTimeout(PingRequest request) {
         System.out.println("Got timeout for address " + request.getAddress().getHostAddress()+ " id: " + request.getMessageId());
         counter.countDown();
         responsesNumber += 1;
      }
      
      public void handleException(PingRequest request, Exception exception) {
         System.out.println("Got exception for address " + request.getAddress().getHostAddress() + " id: " + request.getMessageId() + " " + exception.toString());
         counter.countDown();
         responsesNumber += 1;
      }

      public int getResponsesNumber() {
         return responsesNumber;
      }
      
   }
   
   
   /**
    * Create the test case
    *
    * @param testName name of the test case
    */
   public PingTest( String testName )
   {
      super( testName );
      destinationAddresses = new InetAddress[testHosts.length];
      
      for(int i = 0; i < testHosts.length; i++)
         try {
         destinationAddresses[i] = Inet4Address.getByName(testHosts[i]);
      } catch (UnknownHostException ex) {
         System.out.println("Unknown host: " + testHosts[i] + " - ignored");
      }
      return;
   }
   
   /**
    * @return the suite of tests being tested
    */
   public static Test suite()
   {
      return new TestSuite( PingTest.class );
   }
   
   /**
    * Test Ping initialization/release operations.
    */
   public void testPingInit()
   {
      try {
         Ping.init();
         System.out.println("Ping initialized");
         assertTrue( true );
         Ping.release();
         System.out.println("Ping closed");
         assertTrue( true );
      } catch (PingErrorException ex) {
         ex.printStackTrace();
         assertTrue( false );
      }
      
   }
   
   /**
    * Test Ping echo request.
    * Test waits for response 1 second asuming it is enoush for ping.
    */
   public void testPingEcho()
   {
      try
      {
         CountDownLatch counter = new CountDownLatch(1);
         ResponseHandler handler = new ResponseHandler(counter);
         Ping.init();
         System.out.println("PingEngine initialized");
         assertTrue( true );
         Ping.echo(Inet4Address.getByName("8.8.8.8"), testData.getBytes(), 5000, handler);
         try {
            Thread.sleep(1000);
         } catch (InterruptedException ex) {
            Logger.getLogger(PingTest.class.getName()).log(Level.SEVERE, null, ex);
         }
         Ping.release();
         System.out.println("PingEngine closed");
         assertTrue(handler.getResponsesNumber() == 1);
      } catch (UnknownHostException ex) {
         System.out.println("testPingEngineEcho - unkbown host");
         assertTrue( false );
      } catch (PingErrorException ex) {
         ex.printStackTrace();
         assertTrue( false );
      }
   }
      
   /**
    * Test Ping echo requests using predefined set of hosts.
    */
   public void testPingEchoMulti()
   {
      try {
         CountDownLatch counter = new CountDownLatch(destinationAddresses.length);
         ResponseHandler handler = new ResponseHandler(counter);

         Ping.init();
         System.out.println("Ping initialized");
         assertTrue( true );
         for(int i = 0; i < destinationAddresses.length; i++)
         {
            Ping.echo(destinationAddresses[i], "TestData".getBytes(), 1000, handler);
         };
         try 
         {
            // Success when counter reaches 0 before timeout
            assertTrue(counter.await(1100, TimeUnit.MILLISECONDS));
         }
         catch (InterruptedException ex) 
         {
            System.out.println("Test interrupted");
         }
         finally
         {
        	Ping.release();
         }
         System.out.println("Ping closed");
         assertTrue(handler.getResponsesNumber() == destinationAddresses.length);
      }
      catch (PingErrorException ex) 
      {
         ex.printStackTrace();
         assertTrue( false );
      }
      
   }
   
   
   /**
    * Test Ping echo requests using predefined set of hosts.
    */
   public void testPingEchoMultiDefault()
   {
      try {
         CountDownLatch counter = new CountDownLatch(destinationAddresses.length);
         ResponseHandler handler = new ResponseHandler(counter);

         Ping.init();
         System.out.println("Ping initialized");
         assertTrue( true );
         for(int i = 0; i < destinationAddresses.length; i++)
         {
            Ping.echo(destinationAddresses[i], testDefaultDataLen, 1000, handler);
         };
         try 
         {
            // Success when counter reaches 0 before timeout
            assertTrue(counter.await(1100, TimeUnit.MILLISECONDS));
         }
         catch (InterruptedException ex) 
         {
            System.out.println("Test interrupted");
         }
         finally
         {
        	Ping.release();
         }
         System.out.println("Ping closed");
         assertTrue(handler.getResponsesNumber() == destinationAddresses.length);
      }
      catch (PingErrorException ex) 
      {
         ex.printStackTrace();
         assertTrue( false );
      }
      
   }
   
   
   /**
    * Test Ping echo requests using predefined set of hosts.
    */
   public void testPingBlocking()
   {
      try {
         PingResult[] result;
         
         Ping.init();
         System.out.println("PingEngine initialized");
         assertTrue( true );
         
         result = Ping.echo(destinationAddresses, testData.getBytes(), 5000);

         for(int i = 0; i < testHosts.length; i++)
            System.out.println("Ping:" + result[i].getAddress().getHostAddress() + "  time:" + result[i].getResponseTime() + "  result:" + result[i].getResult() + "  status:" + result[i].getStatus());
         
         Ping.release();
         System.out.println("PingEngine closed");
      }
      catch (PingErrorException ex) 
      {
         ex.printStackTrace();
         assertTrue( false );
      }
   }
   
   /**
    * Test Ping echo requests using predefined set of hosts.
    */
   public void testPingBlockingDefault()
   {
      try {
         PingResult[] result;
         
         Ping.init();
         System.out.println("PingEngine initialized");
         assertTrue( true );
         
         result = Ping.echo(destinationAddresses, testDefaultDataLen, 5000);

         for(int i = 0; i < testHosts.length; i++)
            System.out.println("Ping:" + result[i].getAddress().getHostAddress() + "  time:" + result[i].getResponseTime() + "  result:" + result[i].getResult() + "  status:" + result[i].getStatus());
         
         Ping.release();
         System.out.println("PingEngine closed");
      }
      catch (PingErrorException ex) 
      {
         ex.printStackTrace();
         assertTrue( false );
      }
   }
   
}
